from __future__ import annotations

import re
import uuid as std_uuid
from typing import Annotated, Any

from fastapi import Depends, HTTPException, status
from fastapi.security import OpenIdConnect
from joserfc.errors import JoseError
from joserfc.jwt import JWTClaimsRegistry
from pydantic import BaseModel, GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import CoreSchema, core_schema
from uuid_utils import UUID as _UUID

from diracx.core.models import UserInfo
from diracx.core.properties import SecurityProperty
from diracx.logic.auth.utils import read_token
from diracx.routers.dependencies import AuthSettings

# auto_error=False is used to avoid raising the wrong exception when the token is missing
# The error is handled in the verify_dirac_access_token function
# More info:
# - https://github.com/tiangolo/fastapi/issues/10177
# - https://datatracker.ietf.org/doc/html/rfc6750#section-3.1
oidc_scheme = OpenIdConnect(
    openIdConnectUrl="/.well-known/openid-configuration", auto_error=False
)


class UUID(_UUID):
    """Subclass of uuid_utils.UUID to add pydantic support."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Use the stdlib uuid.UUID schema for validation and serialization."""
        std_schema = handler(std_uuid.UUID)

        def to_uuid_utils(u: std_uuid.UUID) -> UUID:
            return cls(str(u))

        return core_schema.no_info_after_validator_function(to_uuid_utils, std_schema)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        """Return the stdlib uuid.UUID schema for JSON serialization."""
        return handler(core_schema)


class AuthInfo(BaseModel):
    # raw token for propagation
    bearer_token: str

    # token ID in the DB for Component
    # unique jwt identifier for user
    token_id: UUID

    # list of DIRAC properties
    properties: list[SecurityProperty]

    policies: dict[str, Any] = {}


class AuthorizedUserInfo(AuthInfo, UserInfo):
    pass


async def verify_dirac_access_token(
    authorization: Annotated[str, Depends(oidc_scheme)],
    settings: AuthSettings,
) -> AuthorizedUserInfo:
    """Verify dirac user token and return a UserInfo class
    Used for each API endpoint.
    """
    if not authorization:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Authorization header is missing",
            headers={"WWW-Authenticate": "Bearer"},
        )
    if match := re.fullmatch(r"Bearer (.+)", authorization):
        raw_token = match.group(1)
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Invalid authorization header",
        )

    try:
        claims = read_token(
            raw_token,
            settings.token_keystore.jwks,
            settings.token_allowed_algorithms,
            claims_requests=JWTClaimsRegistry(
                iss={"essential": True, "value": settings.token_issuer},
            ),
        )
    except JoseError as e:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid JWT",
        ) from e

    return AuthorizedUserInfo(
        bearer_token=raw_token,
        token_id=claims["jti"],
        properties=claims["dirac_properties"],
        sub=claims["sub"],
        preferred_username=claims["preferred_username"],
        dirac_group=claims["dirac_group"],
        vo=claims["vo"],
        policies=claims.get("dirac_policies", {}),
    )



from __future__ import annotations

import asyncio
import contextlib
from typing import Any, Callable, TypeVar, cast

from fastapi import APIRouter, FastAPI
from starlette.routing import Route

T = TypeVar("T")


def _downgrade_openapi_schema(data):
    """Modify an openapi schema in-place to be compatible with AutoRest."""
    if isinstance(data, dict):
        for k, v in list(data.items()):
            if k == "anyOf":
                if {"type": "null"} in v:
                    v.pop(v.index({"type": "null"}))
                    data["nullable"] = True
                    if len(v) == 1:
                        data |= v[0]
            elif k == "const":
                data.pop(k)
            # https://github.com/fastapi/fastapi/discussions/12984
            elif k == "propertyNames":
                data.pop(k)

            _downgrade_openapi_schema(v)
    if isinstance(data, list):
        for v in data:
            _downgrade_openapi_schema(v)


class DiracFastAPI(FastAPI):
    def __init__(self):
        @contextlib.asynccontextmanager
        async def lifespan(app: DiracFastAPI):
            async with contextlib.AsyncExitStack() as stack:
                await asyncio.gather(
                    *(stack.enter_async_context(f()) for f in app.lifetime_functions)
                )
                yield

        self.lifetime_functions = []
        super().__init__(
            swagger_ui_init_oauth={
                "clientId": "myDIRACClientID",
                "scopes": "property:NormalUser",
                "usePkceWithAuthorizationCodeGrant": True,
            },
            generate_unique_id_function=lambda route: f"{route.tags[0]}_{route.name}",
            title="Dirac",
            lifespan=lifespan,
            openapi_url="/api/openapi.json",
            docs_url="/api/docs",
            swagger_ui_oauth2_redirect_url="/api/docs/oauth2-redirect",
        )
        # FIXME: when autorest will support 3.1.0
        # From 0.99.0, FastAPI is using openapi 3.1.0 by default
        # This version is not supported by autorest yet
        self.openapi_version = "3.0.2"

    def openapi(self, *args, **kwargs):
        if not self.openapi_schema:
            super().openapi(*args, **kwargs)
            _downgrade_openapi_schema(self.openapi_schema)

            # Remove 422 responses as we don't want autorest to use it
            for _, method_item in self.openapi_schema.get("paths").items():
                for _, param in method_item.items():
                    responses = param.get("responses")
                    if "422" in responses:
                        del responses["422"]

        return self.openapi_schema


class DiracxRouter(APIRouter):
    def __init__(
        self,
        *,
        dependencies=None,
        require_auth: bool = True,
        include_in_schema: bool = True,
        path_root: str = "/api",
    ):
        super().__init__(dependencies=dependencies, include_in_schema=include_in_schema)
        self.diracx_require_auth = require_auth
        self.diracx_path_root = path_root

    ####
    # These 2 methods are needed to overwrite routes
    # https://github.com/tiangolo/fastapi/discussions/8489

    def add_api_route(self, path: str, endpoint: Callable[..., Any], **kwargs):
        route_index = self._get_route_index_by_path_and_methods(
            path, set(kwargs.get("methods", []))
        )
        if route_index >= 0:
            self.routes.pop(route_index)

        return super().add_api_route(path, endpoint, **kwargs)

    def _get_route_index_by_path_and_methods(self, path: str, methods: set[str]) -> int:
        routes = cast(list[Route], self.routes)
        for index, route in enumerate(routes):
            if route.path == path and methods == route.methods:
                return index
        return -1

    ######



"""Authorization code flow.

See docs/admin/explanations/authentication.md
"""

from __future__ import annotations

from typing import Literal

from fastapi import (
    HTTPException,
    Request,
    responses,
    status,
)

from diracx.core.exceptions import AuthorizationError, IAMClientError, IAMServerError
from diracx.logic.auth.authorize_code_flow import (
    complete_authorization_flow as complete_authorization_flow_bl,
)
from diracx.logic.auth.authorize_code_flow import (
    initiate_authorization_flow as initiate_authorization_flow_bl,
)

from ..dependencies import (
    AuthDB,
    AuthSettings,
    AvailableSecurityProperties,
    Config,
)
from ..fastapi_classes import DiracxRouter

router = DiracxRouter(require_auth=False)


@router.get("/authorize")
async def initiate_authorization_flow(
    request: Request,
    response_type: Literal["code"],
    code_challenge: str,
    code_challenge_method: Literal["S256"],
    client_id: str,
    redirect_uri: str,
    scope: str,
    state: str,
    auth_db: AuthDB,
    config: Config,
    available_properties: AvailableSecurityProperties,
    settings: AuthSettings,
) -> responses.RedirectResponse:
    """Initiate the authorization flow.
    It will redirect to the actual OpenID server (IAM, CheckIn) to
    perform a authorization code flow.

    Scope details:
    - If only VO is provided: Uses the default group and its properties for the VO.

    - If VO and group are provided: Uses the specified group and its properties for the VO.

    - If VO and properties are provided: Uses the default group and combines its properties with the
      provided properties.

    - If VO, group, and properties are provided: Uses the specified group and combines its properties with the
      provided properties.

    We set the user details obtained from the user authorize flow in a cookie
    to be able to map the authorization flow with the corresponding
    user authorize flow.
    """
    try:
        redirect_uri = await initiate_authorization_flow_bl(
            request_url=f"{request.url.replace(query='')}",
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method,
            client_id=client_id,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
            auth_db=auth_db,
            config=config,
            settings=settings,
            available_properties=available_properties,
        )
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=str(e),
        ) from e

    return responses.RedirectResponse(redirect_uri)


@router.get("/authorize/complete")
async def complete_authorization_flow(
    code: str,
    state: str,
    request: Request,
    auth_db: AuthDB,
    config: Config,
    settings: AuthSettings,
) -> responses.RedirectResponse:
    """Complete the authorization flow.

    The user is redirected back to the DIRAC auth service after completing the IAM's authorization flow.
    We retrieve the original flow details from the decrypted state and store the ID token requested from the IAM.
    The user is then redirected to the client's redirect URI.
    """
    try:
        redirect_uri = await complete_authorization_flow_bl(
            code=code,
            state=state,
            request_url=str(request.url.replace(query="")),
            auth_db=auth_db,
            config=config,
            settings=settings,
        )
    except AuthorizationError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid state"
        ) from e
    except IAMServerError as e:
        raise HTTPException(
            status_code=status.HTTP_502_BAD_GATEWAY,
            detail="Failed to contact IAM server",
        ) from e
    except IAMClientError as e:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid code"
        ) from e
    return responses.RedirectResponse(redirect_uri)



from __future__ import annotations

from datetime import datetime, timezone
from http import HTTPStatus

import pytest
from fastapi.testclient import TestClient
from freezegun import freeze_time

from diracx.core.models import JobStatus
from diracx.routers.jobs.query import EXAMPLE_SUMMARY

from .conftest import TEST_JDL, TEST_PARAMETRIC_JDL

TEST_LARGE_PARAMETRIC_JDL = """
    Executable = "echo";
    Arguments = "%s";
    JobName = "Test_%n";
    Parameters = 100;
    ParameterStart = 1;
"""

TEST_MALFORMED_JDL = """
[
    'Executable = "echo";'
]
"""

pytestmark = pytest.mark.enabled_dependencies(
    [
        "AuthSettings",
        "JobDB",
        "JobLoggingDB",
        "ConfigSource",
        "TaskQueueDB",
        "SandboxMetadataDB",
        "WMSAccessPolicy",
        "DevelopmentSettings",
        "JobParametersDB",
    ]
)


def test_insert_and_list_parametric_jobs(normal_user_client):
    job_definitions = [TEST_PARAMETRIC_JDL]
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 3  # Parameters.JOB_ID is 3

    submitted_job_ids = sorted([job_dict["JobID"] for job_dict in r.json()])

    r = normal_user_client.post("/api/jobs/search")
    assert r.status_code == 200, r.json()

    listed_jobs = r.json()

    assert "Content-Range" not in r.headers

    assert len(listed_jobs) == 3  # Parameters.JOB_ID is 3

    assert submitted_job_ids == sorted([job_dict["JobID"] for job_dict in listed_jobs])


@pytest.mark.parametrize(
    "job_definitions",
    [
        [TEST_JDL],
        [TEST_JDL for _ in range(2)],
        [TEST_JDL for _ in range(10)],
    ],
)
def test_insert_and_list_bulk_jobs(job_definitions, normal_user_client):
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert r.status_code == 200, r.json()
    assert len(r.json()) == len(job_definitions)

    submitted_job_ids = sorted([job_dict["JobID"] for job_dict in r.json()])

    r = normal_user_client.post("/api/jobs/search")
    assert r.status_code == 200, r.json()

    listed_jobs = r.json()

    assert "Content-Range" not in r.headers

    assert len(listed_jobs) == len(job_definitions)

    assert submitted_job_ids == sorted([job_dict["JobID"] for job_dict in listed_jobs])


def test_insert_and_search(normal_user_client):
    """Test inserting a job and then searching for it."""
    # job_definitions = [TEST_JDL%(normal_user_client.dirac_token_payload)]
    job_definitions = [TEST_JDL]
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == len(job_definitions)

    submitted_job_ids = sorted([job_dict["JobID"] for job_dict in r.json()])

    # Test /jobs/search
    # 1. Search for all jobs
    r = normal_user_client.post("/api/jobs/search")
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert [x["JobID"] for x in listed_jobs] == submitted_job_ids
    assert {x["VerifiedFlag"] for x in listed_jobs} == {True}

    # 2. Search for all jobs with status NEW: should return an empty list
    r = normal_user_client.post(
        "/api/jobs/search",
        json={"search": [{"parameter": "Status", "operator": "eq", "value": "NEW"}]},
    )
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert listed_jobs == []

    assert "Content-Range" not in r.headers

    # 3. Search for all jobs with status RECEIVED: should return the submitted jobs
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "Status",
                    "operator": "eq",
                    "value": JobStatus.RECEIVED.value,
                }
            ]
        },
    )
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert [x["JobID"] for x in listed_jobs] == submitted_job_ids

    assert "Content-Range" not in r.headers

    # 4. Search for all jobs but just return the JobID and the Status
    r = normal_user_client.post(
        "/api/jobs/search", json={"parameters": ["JobID", "Status"]}
    )
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert listed_jobs == [
        {"JobID": jid, "Status": JobStatus.RECEIVED.value} for jid in submitted_job_ids
    ]

    assert "Content-Range" not in r.headers

    # Test /jobs/summary
    r = normal_user_client.post(
        "/api/jobs/summary", json={"grouping": ["Status", "OwnerGroup"]}
    )
    assert r.status_code == 200, r.json()

    assert r.json() == [
        {"Status": JobStatus.RECEIVED.value, "OwnerGroup": "test_group", "count": 1}
    ]

    r = normal_user_client.post(
        "/api/jobs/summary",
        json={
            "grouping": ["Status"],
            "search": [
                {
                    "parameter": "Status",
                    "operator": "eq",
                    "value": JobStatus.RECEIVED.value,
                }
            ],
        },
    )
    assert r.status_code == 200, r.json()
    assert r.json() == [{"Status": JobStatus.RECEIVED.value, "count": 1}]

    r = normal_user_client.post(
        "/api/jobs/summary",
        json={
            "grouping": ["Status"],
            "search": [{"parameter": "Status", "operator": "eq", "value": "NEW"}],
        },
    )
    assert r.status_code == 200, r.json()
    assert r.json() == []


def test_insert_malformed_jdl(normal_user_client):
    job_definitions = [TEST_MALFORMED_JDL]
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert r.status_code == 400, r.json()


# "2024-01-01 00:00:00.123456"
@freeze_time(
    datetime(
        year=2024,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
        microsecond=123456,
        tzinfo=timezone.utc,
    )
)
def test_insert_and_search_by_datetime(normal_user_client):
    """Test inserting a job and then searching for it.

    Focus on the SubmissionTime parameter.
    """
    # job_definitions = [TEST_JDL%(normal_user_client.dirac_token_payload)]
    job_definitions = [TEST_JDL]
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == len(job_definitions)
    r = normal_user_client.post("/api/jobs/search")
    assert len(r.json()) == 1, "No jobs submitted"

    submitted_jobs_info = r.json()

    # 1.1 Search for all jobs submitted in 2024
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1, f"submitted jobs were: {submitted_jobs_info!r}"

    # 1.2 Search for all jobs submitted before 2024
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "lt",
                    "value": "2024",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 0

    # 2.1 Search for all jobs submitted after 2024-01
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "gt",
                    "value": "2024-01",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 0

    # 2.2 Search for all jobs submitted before 2024-02
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "lt",
                    "value": "2024-02",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 3 Search for all jobs submitted during 2024-01-01
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 4.1 Search for all jobs submitted during 2024-01-01 00
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01 00",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 4.2 Search for all jobs submitted during 2024-01-01T00 (with the 'T' separator)
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01T00",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 4.3 Search for all jobs not submitted during 2024-01-01 01
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "neq",
                    "value": "2024-01-01 01",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 5.1 Search for all jobs submitted after 2024-01-01 00:00:00
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "gt",
                    "value": "2024-01-01 00:00:00",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 0

    # 5.2 Search for all jobs not submitted on 2024-01-01 00:00:00
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "neq",
                    "value": "2024-01-01 00:00:00",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 0

    # 5.3 Search for all jobs submitted on 2024-01-01 00:00:00
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01 00:00:00",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 6.1 Search for all jobs submitted on 2024-01-01 00:00:00.123456
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01 00:00:00.123456",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 6.2 Search for all jobs submitted on 2024-01-01 00:00:00.123456Z
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01 00:00:00.123456Z",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1

    # 6.3 Search for all jobs submitted on 2024-01-01 00:00:00.123Z
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "SubmissionTime",
                    "operator": "eq",
                    "value": "2024-01-01 00:00:00.123Z",
                }
            ]
        },
    )
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1


def test_search_distinct(normal_user_client):
    """Test that the distinct parameter works as expected."""
    job_definitions = [TEST_JDL, TEST_JDL, TEST_JDL]
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == len(job_definitions)

    # Check that distinct collapses identical records when true
    r = normal_user_client.post(
        "/api/jobs/search", json={"parameters": ["Status"], "distinct": False}
    )
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) > 1

    assert "Content-Range" not in r.headers

    r = normal_user_client.post(
        "/api/jobs/search", json={"parameters": ["Status"], "distinct": True}
    )
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == 1

    assert "Content-Range" not in r.headers


def test_search_pagination(normal_user_client):
    """Test that the pagination works as expected."""
    job_definitions = [TEST_JDL] * 20
    r = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == len(job_definitions)

    # Get the first 20 jobs (all of them)
    r = normal_user_client.post("/api/jobs/search", params={"page": 1, "per_page": 20})
    listed_jobs = r.json()
    assert r.status_code == 200, listed_jobs
    assert len(listed_jobs) == 20

    assert "Content-Range" not in r.headers

    # Get the first 10 jobs
    r = normal_user_client.post("/api/jobs/search", params={"page": 1, "per_page": 10})
    listed_jobs = r.json()
    assert r.status_code == 206, listed_jobs
    assert len(listed_jobs) == 10

    assert "Content-Range" in r.headers
    assert (
        r.headers["Content-Range"]
        == f"jobs 0-{len(listed_jobs) - 1}/{len(job_definitions)}"
    )

    # Get the next 10 jobs
    r = normal_user_client.post("/api/jobs/search", params={"page": 2, "per_page": 10})
    listed_jobs = r.json()
    assert r.status_code == 206, listed_jobs
    assert len(listed_jobs) == 10

    assert "Content-Range" in r.headers
    assert (
        r.headers["Content-Range"]
        == f"jobs 10-{len(listed_jobs) + 10 - 1}/{len(job_definitions)}"
    )

    # Get an unknown page
    r = normal_user_client.post("/api/jobs/search", params={"page": 3, "per_page": 10})
    listed_jobs = r.json()
    assert r.status_code == 416, listed_jobs
    assert len(listed_jobs) == 0

    assert "Content-Range" in r.headers
    assert r.headers["Content-Range"] == f"jobs */{len(job_definitions)}"

    # Set the per_page parameter to 0
    r = normal_user_client.post("/api/jobs/search", params={"page": 1, "per_page": 0})
    assert r.status_code == 400, r.json()

    # Set the per_page parameter to a negative number
    r = normal_user_client.post("/api/jobs/search", params={"page": 1, "per_page": -1})
    assert r.status_code == 400, r.json()

    # Set the page parameter to 0
    r = normal_user_client.post("/api/jobs/search", params={"page": 0, "per_page": 10})
    assert r.status_code == 400, r.json()

    # Set the page parameter to a negative number
    r = normal_user_client.post("/api/jobs/search", params={"page": -1, "per_page": 10})
    assert r.status_code == 400, r.json()


def test_user_cannot_submit_parametric_jdl_greater_than_max_parametric_jobs(
    normal_user_client,
):
    """Test that a user cannot submit a parametric JDL greater than the max parametric jobs."""
    job_definitions = [TEST_LARGE_PARAMETRIC_JDL]
    res = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert res.status_code == HTTPStatus.BAD_REQUEST, res.json()


def test_user_cannot_submit_list_of_jdl_greater_than_max_number_of_jobs(
    normal_user_client,
):
    """Test that a user cannot submit a list of JDL greater than the max number of jobs."""
    job_definitions = [TEST_JDL for _ in range(100)]
    res = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert res.status_code == HTTPStatus.BAD_REQUEST, res.json()


@pytest.mark.parametrize(
    "job_definitions",
    [[TEST_PARAMETRIC_JDL, TEST_JDL], [TEST_PARAMETRIC_JDL, TEST_PARAMETRIC_JDL]],
)
def test_user_cannot_submit_multiple_jdl_if_at_least_one_of_them_is_parametric(
    normal_user_client, job_definitions
):
    res = normal_user_client.post("/api/jobs/jdl", json=job_definitions)
    assert res.status_code == HTTPStatus.BAD_REQUEST, res.json()


def test_user_without_the_normal_user_property_cannot_submit_job(admin_user_client):
    pytest.skip(
        "AlwaysAllowAccessPolicyCallable is forced in testing, so this test can not actually test this access policy."
    )
    res = admin_user_client.post("/api/jobs/jdl", json=[TEST_JDL])
    assert res.status_code == HTTPStatus.FORBIDDEN, res.json()


def test_get_job_status(normal_user_client: TestClient, valid_job_id: int):
    """Test that the job status is returned correctly."""
    # Act
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "parameters": ["JobID", "Status", "MinorStatus", "ApplicationStatus"],
            "search": [{"parameter": "JobID", "operator": "eq", "value": valid_job_id}],
        },
    )

    # Assert
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1, f"Should only return length-1 list: {r.json()}"
    assert r.json()[0]["JobID"] == valid_job_id, "Returned wrong job id"
    # TODO: should we return camel case here (and everywhere else) ?
    assert r.json()[0]["Status"] == JobStatus.RECEIVED.value
    assert r.json()[0]["MinorStatus"] == "Job accepted"
    assert r.json()[0]["ApplicationStatus"] == "Unknown"


def test_get_status_of_nonexistent_job(
    normal_user_client: TestClient, invalid_job_id: int
):
    """Test that the job status is returned correctly."""
    # Act
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "parameters": ["Status"],
            "search": [
                {"parameter": "JobID", "operator": "eq", "value": invalid_job_id}
            ],
        },
    )

    # Assert
    assert r.status_code == HTTPStatus.OK, r.json()
    assert r.json() == []


def test_get_job_status_in_bulk(normal_user_client: TestClient, valid_job_ids: list):
    """Test that we can get the status of multiple jobs in one request."""
    # Act

    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "parameters": ["JobID", "Status", "MinorStatus", "ApplicationStatus"],
            "search": [
                {"parameter": "JobID", "operator": "in", "values": valid_job_ids}
            ],
        },
    )

    # Assert
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 3  # Parameters.JOB_ID is 3
    assert {j["JobID"] for j in r.json()} == set(valid_job_ids)
    for job in r.json():
        assert job["JobID"] in valid_job_ids
        assert job["Status"] == JobStatus.SUBMITTING.value
        assert job["MinorStatus"] == "Bulk transaction confirmation"
        assert job["ApplicationStatus"] == "Unknown"


async def test_get_job_status_history(
    normal_user_client: TestClient, valid_job_id: int
):
    # Arrange
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "parameters": ["JobID", "Status", "MinorStatus", "ApplicationStatus"],
            "search": [{"parameter": "JobID", "operator": "eq", "value": valid_job_id}],
        },
    )
    assert r.status_code == 200, r.json()
    assert r.json()[0]["Status"] == JobStatus.RECEIVED.value
    assert r.json()[0]["MinorStatus"] == "Job accepted"
    assert r.json()[0]["ApplicationStatus"] == "Unknown"

    new_status = JobStatus.CHECKING.value
    new_minor_status = "JobPath"
    before = datetime.now(timezone.utc)

    r = normal_user_client.patch(
        "/api/jobs/status",
        json={
            valid_job_id: {
                datetime.now(tz=timezone.utc).isoformat(): {
                    "Status": new_status,
                    "MinorStatus": new_minor_status,
                }
            }
        },
    )

    after = datetime.now(timezone.utc)

    assert r.status_code == 200, r.json()
    assert r.json()["success"][str(valid_job_id)]["Status"] == new_status
    assert r.json()["success"][str(valid_job_id)]["MinorStatus"] == new_minor_status

    # Act
    r = normal_user_client.post(
        "/api/jobs/search",
        json={
            "parameters": [
                "JobID",
                "Status",
                "MinorStatus",
                "ApplicationStatus",
                "LoggingInfo",
            ],
            "search": [{"parameter": "JobID", "operator": "eq", "value": valid_job_id}],
        },
    )
    # Assert
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1
    assert len(r.json()[0]["LoggingInfo"]) == 2
    assert r.json()[0]["LoggingInfo"][0]["Status"] == JobStatus.RECEIVED.value
    assert r.json()[0]["LoggingInfo"][0]["MinorStatus"] == "Job accepted"
    assert r.json()[0]["LoggingInfo"][0]["ApplicationStatus"] == "Unknown"
    assert r.json()[0]["LoggingInfo"][0]["Source"] == "JobManager"

    assert r.json()[0]["LoggingInfo"][1]["Status"] == JobStatus.CHECKING.value
    assert r.json()[0]["LoggingInfo"][1]["MinorStatus"] == "JobPath"
    assert r.json()[0]["LoggingInfo"][1]["ApplicationStatus"] == "Unknown"
    assert (
        before
        < datetime.fromisoformat(r.json()[0]["LoggingInfo"][1]["StatusTime"])
        < after
    )
    assert r.json()[0]["LoggingInfo"][1]["Source"] == "Unknown"


# Test setting job properties


def test_set_single_job_properties(normal_user_client: TestClient, valid_job_id: int):
    pytest.skip("There seems to be a missing route for this - TODO")

    job_id = str(valid_job_id)

    initial_job_state = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "JobID",
                    "operator": "eq",
                    "value": job_id,
                }
            ]
        },
    ).json()[0]

    initial_user_priority = initial_job_state["UserPriority"]
    initial_application_status = initial_job_state["ApplicationStatus"]
    initial_last_update_time = initial_job_state["LastUpdateTime"]

    # Update just one property
    res = normal_user_client.patch(
        "/api/jobs/",
        json={valid_job_id: {"UserPriority": 2}},
    )
    assert res.status_code == 200, res.json()

    new_job_state = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "JobID",
                    "operator": "eq",
                    "value": job_id,
                }
            ]
        },
    ).json()[0]
    new_user_priority = new_job_state["UserPriority"]
    new_application_status = new_job_state["ApplicationStatus"]

    assert initial_application_status == new_application_status
    assert initial_user_priority != new_user_priority
    assert new_user_priority == 2
    assert new_job_state["LastUpdateTime"] == initial_last_update_time

    # Update two properties
    res = normal_user_client.patch(
        f"/api/jobs/{job_id}",
        json={"UserPriority": initial_user_priority, "ApplicationStatus": "Crack"},
        params={"update_timestamp": True},
    )
    assert res.status_code == 200, res.json()

    new_job_state = normal_user_client.post(
        "/api/jobs/search",
        json={
            "search": [
                {
                    "parameter": "JobID",
                    "operator": "eq",
                    "value": job_id,
                }
            ]
        },
    ).json()[0]
    new_user_priority = new_job_state["UserPriority"]
    new_application_status = new_job_state["ApplicationStatus"]

    assert initial_application_status != new_application_status
    assert new_application_status == "Crack"
    assert initial_user_priority == new_user_priority
    assert new_job_state["LastUpdateTime"] != initial_last_update_time


@pytest.mark.parametrize(
    "column_name,expected_initial,updated_values",
    [
        ("AccountedFlag", False, [True, False, "Failed", True, "Failed"]),
        # "New"-style jobs are always inserted with VerifiedFlag=True, no idea why...
        ("VerifiedFlag", True, [True, False, True]),
    ],
)
def test_setting_flag(
    normal_user_client: TestClient,
    valid_job_id: int,
    column_name: str,
    expected_initial: bool | str,
    updated_values: list[bool | str],
):
    search_query = {
        "search": [
            {
                "parameter": "JobID",
                "operator": "eq",
                "value": valid_job_id,
            }
        ]
    }

    res = normal_user_client.post("/api/jobs/search", json=search_query)
    assert res.status_code == 200, res.text
    assert len(res.json()) == 1
    assert res.json()[0][column_name] == expected_initial

    for update in updated_values:
        res = normal_user_client.patch(
            "/api/jobs/metadata", json={valid_job_id: {column_name: update}}
        )
        assert res.status_code == 204, res.text

        res = normal_user_client.post("/api/jobs/search", json=search_query)
        assert res.status_code == 200, res.text
        assert len(res.json()) == 1
        assert res.json()[0][column_name] == update


def test_set_single_job_properties_non_existing_job(
    normal_user_client: TestClient, invalid_job_id: int
):
    pytest.skip("There seems to be a missing route for this - TODO")
    job_id = str(invalid_job_id)

    res = normal_user_client.patch(
        "/api/jobs/",
        json={job_id: {"UserPriority": 2}},
    )
    assert res.status_code == HTTPStatus.NOT_FOUND, res.json()


def test_get_job_status_history_in_bulk(
    normal_user_client: TestClient, valid_job_id: int
):
    pytest.skip("TODO: decide whether to keep this")

    # Arrange
    r = normal_user_client.get(f"/api/jobs/{valid_job_id}/status")
    assert r.status_code == 200, r.json()
    assert r.json()[str(valid_job_id)]["Status"] == JobStatus.RECEIVED.value
    assert r.json()[str(valid_job_id)]["MinorStatus"] == "Job accepted"
    assert r.json()[str(valid_job_id)]["ApplicationStatus"] == "Unknown"

    # Act
    r = normal_user_client.get(
        "/api/jobs/status/history", params={"job_ids": [valid_job_id]}
    )

    # Assert
    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1
    assert r.json()[str(valid_job_id)][0]["Status"] == JobStatus.RECEIVED.value
    assert r.json()[str(valid_job_id)][0]["MinorStatus"] == "Job accepted"
    assert r.json()[str(valid_job_id)][0]["ApplicationStatus"] == "Unknown"
    assert datetime.fromisoformat(r.json()[str(valid_job_id)][0]["StatusTime"])
    assert r.json()[str(valid_job_id)][0]["Source"] == "JobManager"


def test_summary(normal_user_client: TestClient, valid_job_id: int):
    """Test that the summary endpoint works as expected."""
    r = normal_user_client.post(
        "/api/jobs/summary",
        json={
            "grouping": ["Status", "OwnerGroup"],
            "search": [{"parameter": "JobID", "operator": "eq", "value": valid_job_id}],
        },
    )
    assert r.status_code == 200, r.json()
    assert r.json() == [
        {"Status": JobStatus.RECEIVED.value, "OwnerGroup": "test_group", "count": 1}
    ]

    r = normal_user_client.post(
        "/api/jobs/summary",
        json={
            "grouping": ["Owner"],
            "search": [],
        },
    )

    assert r.status_code == 200, r.json()
    assert r.json() == [{"Owner": "preferred_username", "count": 1}]


def test_summary_doc_example(normal_user_client: TestClient, valid_job_id: int):
    """Test that the summary doc example is correct."""
    payload = EXAMPLE_SUMMARY["Group by JobGroup"]["value"]
    r = normal_user_client.post("/api/jobs/summary", json=payload)

    assert r.status_code == 200, r.json()
    assert len(r.json()) == 1



from __future__ import annotations

__all__ = ["router"]

from ..fastapi_classes import DiracxRouter
from .probes import router as probes_router

router = DiracxRouter(require_auth=False, include_in_schema=False)
router.include_router(probes_router)



"""Token endpoint."""

from __future__ import annotations

import logging
import os
from typing import Annotated, Literal

from fastapi import Depends, Form, Header, HTTPException, status
from joserfc.errors import JoseError

from diracx.core.exceptions import (
    DiracHttpResponseError,
    InvalidCredentialsError,
    PendingAuthorizationError,
)
from diracx.core.models import (
    AccessTokenPayload,
    GrantType,
    RefreshTokenPayload,
    TokenResponse,
)
from diracx.logic.auth.token import create_token
from diracx.logic.auth.token import get_oidc_token as get_oidc_token_bl
from diracx.logic.auth.token import (
    perform_legacy_exchange as perform_legacy_exchange_bl,
)
from diracx.routers.access_policies import BaseAccessPolicy

from ..dependencies import AuthDB, AuthSettings, AvailableSecurityProperties, Config
from ..fastapi_classes import DiracxRouter

router = DiracxRouter(require_auth=False)
logger = logging.getLogger(__name__)


async def mint_token(
    access_payload: AccessTokenPayload,
    refresh_payload: RefreshTokenPayload | None,
    existing_refresh_token: str | None,
    all_access_policies: dict[str, BaseAccessPolicy],
    settings: AuthSettings,
) -> TokenResponse:
    """Enrich the token with policy specific content and mint it."""
    if not refresh_payload and not existing_refresh_token:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Refresh token is not set and no refresh token was provided",
        )

    # Enrich the token with policy specific content
    dirac_access_policies = {}
    dirac_refresh_policies = {}
    for policy_name, policy in all_access_policies.items():
        access_extra, refresh_extra = policy.enrich_tokens(
            access_payload, refresh_payload
        )
        if access_extra:
            dirac_access_policies[policy_name] = access_extra
        if refresh_extra:
            dirac_refresh_policies[policy_name] = refresh_extra

    # Create the access token
    access_payload["dirac_policies"] = dirac_access_policies
    access_token = create_token(access_payload, settings)

    # Create the refresh token
    if refresh_payload:
        refresh_payload["dirac_policies"] = dirac_refresh_policies
        refresh_token = create_token(refresh_payload, settings)
    elif existing_refresh_token:
        refresh_token = existing_refresh_token

    return TokenResponse(
        access_token=access_token,
        expires_in=settings.access_token_expire_minutes * 60,
        refresh_token=refresh_token,
    )


@router.post("/token")
async def get_oidc_token(
    # Autorest does not support the GrantType annotation
    # We need to specify each option with Literal[]
    grant_type: Annotated[
        Literal[GrantType.authorization_code]
        | Literal[GrantType.device_code]
        | Literal[GrantType.refresh_token],
        Form(description="OAuth2 Grant type"),
    ],
    client_id: Annotated[str, Form(description="OAuth2 client id")],
    auth_db: AuthDB,
    config: Config,
    settings: AuthSettings,
    available_properties: AvailableSecurityProperties,
    all_access_policies: Annotated[
        dict[str, BaseAccessPolicy], Depends(BaseAccessPolicy.all_used_access_policies)
    ],
    device_code: Annotated[
        str | None, Form(description="device code for OAuth2 device flow")
    ] = None,
    code: Annotated[
        str | None, Form(description="Code for OAuth2 authorization code flow")
    ] = None,
    redirect_uri: Annotated[
        str | None,
        Form(description="redirect_uri used with OAuth2 authorization code flow"),
    ] = None,
    code_verifier: Annotated[
        str | None,
        Form(
            description="Verifier for the code challenge for the OAuth2 authorization flow with PKCE"
        ),
    ] = None,
    refresh_token: Annotated[
        str | None,
        Form(description="Refresh token used with OAuth2 refresh token flow"),
    ] = None,
) -> TokenResponse:
    """Token endpoint to retrieve the token at the end of a flow.
    This is the endpoint being pulled by dirac-login when doing the device flow.
    """
    try:
        access_payload, refresh_payload = await get_oidc_token_bl(
            grant_type,
            client_id,
            auth_db,
            config,
            settings,
            available_properties,
            device_code=device_code,
            code=code,
            redirect_uri=redirect_uri,
            code_verifier=code_verifier,
            refresh_token=refresh_token,
        )
    except PendingAuthorizationError as e:
        raise DiracHttpResponseError(
            status_code=status.HTTP_400_BAD_REQUEST,
            data={"error": "authorization_pending"},
        ) from e
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=str(e),
        ) from e
    except (
        InvalidCredentialsError,
        JoseError,
    ) as e:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(e),
            headers={"WWW-Authenticate": "Bearer"},
        ) from e
    except PermissionError as e:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=str(e),
        ) from e
    return await mint_token(
        access_payload, refresh_payload, refresh_token, all_access_policies, settings
    )


BASE_64_URL_SAFE_PATTERN = (
    r"(?:[A-Za-z0-9\-_]{4})*(?:[A-Za-z0-9\-_]{2}==|[A-Za-z0-9\-_]{3}=)?"
)
LEGACY_EXCHANGE_PATTERN = rf"Bearer diracx:legacy:({BASE_64_URL_SAFE_PATTERN})"


@router.get("/legacy-exchange", include_in_schema=False)
async def perform_legacy_exchange(
    preferred_username: str,
    scope: str,
    authorization: Annotated[str, Header()],
    auth_db: AuthDB,
    available_properties: AvailableSecurityProperties,
    settings: AuthSettings,
    config: Config,
    all_access_policies: Annotated[
        dict[str, BaseAccessPolicy], Depends(BaseAccessPolicy.all_used_access_policies)
    ],
    expires_minutes: int | None = None,
) -> TokenResponse:
    """Endpoint used by legacy DIRAC to mint tokens for proxy -> token exchange.

    This route is disabled if DIRACX_LEGACY_EXCHANGE_HASHED_API_KEY is not set
    in the environment.
    """
    if not (
        expected_api_key := os.environ.get("DIRACX_LEGACY_EXCHANGE_HASHED_API_KEY")
    ):
        raise HTTPException(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            detail="Legacy exchange is not enabled",
        )

    try:
        access_payload, refresh_payload = await perform_legacy_exchange_bl(
            expected_api_key=expected_api_key,
            preferred_username=preferred_username,
            scope=scope,
            authorization=authorization,
            auth_db=auth_db,
            available_properties=available_properties,
            settings=settings,
            config=config,
            expires_minutes=expires_minutes,
        )
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=str(e),
        ) from e
    except InvalidCredentialsError as e:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=str(e),
            headers={"WWW-Authenticate": "Bearer"},
        ) from e
    except PermissionError as e:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=str(e),
        ) from e
    return await mint_token(
        access_payload, refresh_payload, None, all_access_policies, settings
    )



from __future__ import annotations

import inspect
from collections import defaultdict
from typing import TYPE_CHECKING

from diracx.core.extensions import select_from_extension
from diracx.routers.access_policies import (
    BaseAccessPolicy,
)

if TYPE_CHECKING:
    from diracx.routers.fastapi_classes import DiracxRouter


def test_all_routes_have_policy():
    """Loop over all the routers, loop over every route,
    and make sure there is a dependency on a BaseAccessPolicy class.

    If the router is created with "require_auth=False", we skip it.
    We also skip routes that have the "diracx_open_access" decorator

    """
    missing_security: defaultdict[list[str]] = defaultdict(list)
    for entry_point in select_from_extension(group="diracx.services"):
        router: DiracxRouter = entry_point.load()

        # If the router was created with the
        # require_auth = False, skip it
        if not router.diracx_require_auth:
            continue

        for route in router.routes:
            # If the route is decorated with the diracx_open_access
            # decorator, we skip it
            if getattr(route.endpoint, "diracx_open_access", False):
                continue

            for dependency in route.dependant.dependencies:
                if inspect.ismethod(dependency.call) and issubclass(
                    dependency.call.__self__, BaseAccessPolicy
                ):
                    # We found a dependency on check_permissions
                    break
            else:
                # We looked at all dependency without finding
                # check_permission
                missing_security[entry_point.name].append(route.name)

    assert not missing_security



from __future__ import annotations

from ..fastapi_classes import DiracxRouter
from ..utils.users import verify_dirac_access_token
from .authorize_code_flow import router as authorize_code_flow_router
from .device_flow import router as device_flow_router
from .management import router as management_router
from .token import router as token_router
from .utils import has_properties

router = DiracxRouter(require_auth=False)
router.include_router(device_flow_router)
router.include_router(management_router)
router.include_router(authorize_code_flow_router)
router.include_router(token_router)

__all__ = ["has_properties", "verify_dirac_access_token"]



from __future__ import annotations  # noqa: I001

import base64
import hashlib
import json
import secrets
from datetime import datetime, timedelta, timezone
from pathlib import Path
from urllib.parse import parse_qs, urlparse

import httpx
import jwt
import pytest
from cryptography.fernet import Fernet
from joserfc.jwk import RSAKey, OKPKey, KeySet
from joserfc.errors import (
    UnsupportedKeyOperationError,
    UnsupportedAlgorithmError,
    InvalidKeyIdError,
)
from pytest_httpx import HTTPXMock
from uuid_utils import uuid7

from diracx.core.config import Config
from diracx.core.exceptions import AuthorizationError
from diracx.core.models import GrantType
from diracx.core.properties import NORMAL_USER, PROXY_MANAGEMENT, SecurityProperty
from diracx.core.settings import AuthSettings
from diracx.logic.auth.token import create_token
from diracx.logic.auth.utils import (
    _server_metadata_cache,
    decrypt_state,
    encrypt_state,
    get_server_metadata,
    parse_and_validate_scope,
    verify_dirac_refresh_token,
)

DIRAC_CLIENT_ID = "myDIRACClientID"
pytestmark = pytest.mark.enabled_dependencies(
    ["AuthDB", "AuthSettings", "ConfigSource", "BaseAccessPolicy"]
)


@pytest.fixture
def test_client(client_factory):
    with client_factory.unauthenticated() as client:
        yield client


@pytest.fixture
def non_mocked_hosts(test_client) -> list[str]:
    return [test_client.base_url.host]


@pytest.fixture
async def auth_httpx_mock(httpx_mock: HTTPXMock, monkeypatch):
    data_dir = Path(__file__).parent.parent / "data"
    path = "idp-server.invalid/.well-known/openid-configuration"
    httpx_mock.add_response(url=f"https://{path}", text=(data_dir / path).read_text())

    # Since 0.32.0, pytest_httpx does not expect to be queried multiple
    # times for the same URL. So force it to allow it
    # By default, it should be done on a per test bases, but well...
    # https://colin-b.github.io/pytest_httpx/#allow-to-register-a-response-for-more-than-one-request
    httpx_mock._options.can_send_already_matched_responses = True

    server_metadata = await get_server_metadata(f"https://{path}")

    id_tokens = ["user1", "user2"]

    def custom_response(request: httpx.Request):
        if b"&code=valid-code&" in request.content:
            id_token = id_tokens.pop(0)
            return httpx.Response(status_code=200, json={"id_token": id_token})
        return httpx.Response(status_code=401)

    httpx_mock.add_callback(custom_response, url=server_metadata["token_endpoint"])

    monkeypatch.setattr("diracx.logic.auth.utils.parse_id_token", fake_parse_id_token)

    yield httpx_mock

    _server_metadata_cache.clear()


async def fake_parse_id_token(raw_id_token: str, *args, **kwargs):
    """Return a fake ID token as if it were returned by an external IdP."""
    id_tokens = {
        "user1": {
            "aud": "5c0541bf-85c8-4d7f-b1df-beaeea19ff5b",
            "email": "christophe.haen@cern.ch",
            "exp": 1680613292,
            "iat": 1680612692,
            "iss": "https://iam-auth.web.cern.ch/",
            "jti": "38dbb060-19ad-4a77-9c54-15901b96e286",
            "kid": "rsa1",
            "name": "Christophe Haen",
            "organisation_name": "lhcb",
            "preferred_username": "chaen",
            "sub": "b824d4dc-1f9d-4ee8-8df5-c0ae55d46041",
            "wlcg.ver": "1.0",
        },
        "user2": {
            "aud": "5c0541bf-85c8-4d7f-b1df-beaeea19ff5b",
            "email": "albdr@email.com",
            "exp": 1680613292,
            "iat": 1680612692,
            "iss": "https://iam-auth.web.cern.ch/",
            "jti": "49ecc171-20be-5b88-0d65-26012c07f397",
            "kid": "rsa1",
            "name": "Albert Durie",
            "organisation_name": "lhcb",
            "preferred_username": "albdr",
            "sub": "c935e5ed-2g0e-5ff9-9eg6-d1bf66e57152",
            "wlcg.ver": "1.0",
        },
    }
    content = id_tokens.get(raw_id_token)
    if not content:
        raise NotImplementedError(raw_id_token)
    return content


async def test_authorization_flow(test_client, auth_httpx_mock: HTTPXMock):
    code_verifier = secrets.token_hex()
    code_challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode()).digest())
        .decode()
        .replace("=", "")
    )

    # The scope is valid and should return a token with the following claims
    # vo:lhcb group:lhcb_user (default group) property:[NormalUser,ProductionManagement]
    # Note: the property ProductionManagement is not part of the lhcb_user group properties
    # but the user has the right to have it.
    scope = "vo:lhcb property:ProductionManagement"

    # Initiate the authorization flow with a wrong client ID
    # Check that the client ID is not recognized
    r = test_client.get(
        "/api/auth/authorize",
        params={
            "response_type": "code",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "client_id": "Unknown client ID",
            "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
            "scope": scope,
            "state": "external-state",
        },
        follow_redirects=False,
    )
    assert r.status_code == 400, r.text

    # Initiate the authorization flow with an unrecognized redirect URI
    # Check that the redirect URI is not recognized
    r = test_client.get(
        "/api/auth/authorize",
        params={
            "response_type": "code",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "client_id": DIRAC_CLIENT_ID,
            "redirect_uri": "http://diracx.test.unrecognized:8000/api/docs/oauth2-redirect",
            "scope": scope,
            "state": "external-state",
        },
        follow_redirects=False,
    )
    assert r.status_code == 400, r.text

    # Correctly initiate the authorization flow
    r = test_client.get(
        "/api/auth/authorize",
        params={
            "response_type": "code",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "client_id": DIRAC_CLIENT_ID,
            "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
            "scope": scope,
            "state": "external-state",
        },
        follow_redirects=False,
    )
    assert r.status_code == 307, r.text
    query_parameters = parse_qs(urlparse(r.headers["Location"]).query)
    redirect_uri = query_parameters["redirect_uri"][0]
    state = query_parameters["state"][0]

    # Check that an invalid code returns an error
    r = test_client.get(redirect_uri, params={"code": "invalid-code", "state": state})
    assert r.status_code == 401, r.text

    # Check that an invalid state returns an error
    r = test_client.get(
        redirect_uri, params={"code": "invalid-code", "state": "invalid-state"}
    )
    assert r.status_code == 400, r.text
    assert "Invalid state" in r.text

    # See if a valid code works
    r = test_client.get(
        redirect_uri,
        params={"code": "valid-code", "state": state},
        follow_redirects=False,
    )
    assert r.status_code == 307, r.text
    assert urlparse(r.headers["Location"]).netloc == "diracx.test.invalid:8000"
    assert urlparse(r.headers["Location"]).path == "/api/docs/oauth2-redirect"
    query_parameters = parse_qs(urlparse(r.headers["Location"]).query)
    assert query_parameters["state"][0] == "external-state"
    code = query_parameters["code"][0]

    # Try to get token with the wrong client ID
    request_data = {
        "grant_type": "authorization_code",
        "code": code,
        "state": state,
        "client_id": "Unknown client ID",
        "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
        "code_verifier": code_verifier,
    }
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 400, r.json()

    # Try to get token with the wrong redirect URI
    request_data = {
        "grant_type": "authorization_code",
        "code": code,
        "state": state,
        "client_id": "Unknown client ID",
        "redirect_uri": "http://diracx.test.unrecognized:8000/api/docs/oauth2-redirect",
        "code_verifier": code_verifier,
    }
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 400, r.json()

    # Get and check token
    request_data = {
        "grant_type": "authorization_code",
        "code": code,
        "state": state,
        "client_id": DIRAC_CLIENT_ID,
        "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
        "code_verifier": code_verifier,
    }
    _get_and_check_token_response(
        test_client,
        request_data=request_data,
    )

    # Ensure the token request doesn't work a second time
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 400, r.json()
    assert r.json()["detail"] == "Code was already used"


async def test_device_flow(test_client, auth_httpx_mock: HTTPXMock):
    # The scope is valid and should return a token with the following claims
    # vo:lhcb group:lhcb_user (default group) property:[NormalUser,ProductionManagement]
    # Note: the property ProductionManagement is not part of the lhcb_user group properties
    # but the user has the right to have it.
    scope = "vo:lhcb property:ProductionManagement"

    # Initiate the device flow with a wrong client ID
    # Check that the client ID is not recognized
    r = test_client.post(
        "/api/auth/device",
        params={
            "client_id": "Unknown client ID",
            "scope": scope,
        },
    )
    assert r.status_code == 400, r.json()

    # Initiate the device flow (would normally be done from CLI)
    r = test_client.post(
        "/api/auth/device",
        params={
            "client_id": DIRAC_CLIENT_ID,
            "scope": scope,
        },
    )
    assert r.status_code == 200, r.json()
    data = r.json()
    assert data["user_code"]
    assert data["device_code"]
    assert data["verification_uri_complete"]
    assert data["verification_uri"]
    assert data["expires_in"] == 600

    # Check that token requests return "authorization_pending"
    r = test_client.post(
        "/api/auth/token",
        data={
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": data["device_code"],
            "client_id": DIRAC_CLIENT_ID,
        },
    )
    assert r.status_code == 400, r.json()
    assert r.json()["error"] == "authorization_pending"

    # Open the DIRAC login page and ensure it redirects to the IdP
    r = test_client.get(data["verification_uri_complete"], follow_redirects=False)
    assert r.status_code == 307, r.text
    login_url = r.headers["Location"]
    assert "/authorize?response_type=code" in login_url
    query_parameters = parse_qs(urlparse(login_url).query)
    redirect_uri = query_parameters["redirect_uri"][0]
    state = query_parameters["state"][0]

    # Check that an invalid code returns an error
    r = test_client.get(redirect_uri, params={"code": "invalid-code", "state": state})
    assert r.status_code == 401, r.text

    # Check that an invalid state returns an error
    r = test_client.get(
        redirect_uri, params={"code": "invalid-code", "state": "invalid-state"}
    )
    assert r.status_code == 400, r.text
    assert "Invalid state" in r.text

    # See if a valid code works
    r = test_client.get(redirect_uri, params={"code": "valid-code", "state": state})
    assert r.status_code == 200, r.text
    assert "Please close the window" in r.text

    # Ensure a valid code does not work a second time
    r = test_client.get(redirect_uri, params={"code": "valid-code", "state": state})
    assert r.status_code == 400, r.text

    # Try to get token with the wrong redirect URI
    request_data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": data["device_code"],
        "client_id": "Unknown client ID",
    }
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 400, r.json()

    # Get and check token
    request_data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": data["device_code"],
        "client_id": DIRAC_CLIENT_ID,
    }
    _get_and_check_token_response(
        test_client,
        request_data=request_data,
    )

    # Ensure the token request doesn't work a second time
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 400, r.json()
    assert r.json()["detail"] == "Code was already used"


async def test_authorization_flow_with_unallowed_properties(
    test_client, auth_httpx_mock: HTTPXMock
):
    """Test the authorization flow and the device flow with unallowed properties."""
    # ProxyManagement is a valid property but not allowed for the user
    unallowed_property = "ProxyManagement"

    # Initiate the authorization flow: should not fail
    code_verifier = secrets.token_hex()
    code_challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode()).digest())
        .decode()
        .replace("=", "")
    )
    r = test_client.get(
        "/api/auth/authorize",
        params={
            "response_type": "code",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "client_id": DIRAC_CLIENT_ID,
            "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
            "scope": f"vo:lhcb property:{unallowed_property} property:NormalUser",
            "state": "external-state",
        },
        follow_redirects=False,
    )
    assert r.status_code == 307, r.json()
    query_parameters = parse_qs(urlparse(r.headers["Location"]).query)
    redirect_uri = query_parameters["redirect_uri"][0]
    state = query_parameters["state"][0]

    r = test_client.get(
        redirect_uri,
        params={"code": "valid-code", "state": state},
        follow_redirects=False,
    )
    assert r.status_code == 307, r.text
    query_parameters = parse_qs(urlparse(r.headers["Location"]).query)
    code = query_parameters["code"][0]

    request_data = {
        "grant_type": "authorization_code",
        "code": code,
        "state": state,
        "client_id": DIRAC_CLIENT_ID,
        "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
        "code_verifier": code_verifier,
    }
    # Ensure the token request doesn't work because of the unallowed property
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 403, r.json()
    assert (
        f"{unallowed_property} are not valid properties for user" in r.json()["detail"]
    )


async def test_device_flow_with_unallowed_properties(
    test_client, auth_httpx_mock: HTTPXMock
):
    """Test the authorization flow and the device flow with unallowed properties."""
    # ProxyManagement is a valid property but not allowed for the user
    unallowed_property = "ProxyManagement"

    # Initiate the device flow
    r = test_client.post(
        "/api/auth/device",
        params={
            "client_id": DIRAC_CLIENT_ID,
            "scope": f"vo:lhcb group:lhcb_user property:{unallowed_property} property:NormalUser",
        },
    )
    assert r.status_code == 200, r.json()

    data = r.json()
    assert data["user_code"]
    assert data["device_code"]
    assert data["verification_uri_complete"]
    assert data["verification_uri"]
    assert data["expires_in"] == 600

    r = test_client.get(data["verification_uri_complete"], follow_redirects=False)
    assert r.status_code == 307, r.text
    login_url = r.headers["Location"]
    query_parameters = parse_qs(urlparse(login_url).query)
    redirect_uri = query_parameters["redirect_uri"][0]
    state = query_parameters["state"][0]

    r = test_client.get(redirect_uri, params={"code": "valid-code", "state": state})
    assert r.status_code == 200, r.text

    request_data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": data["device_code"],
        "client_id": DIRAC_CLIENT_ID,
    }

    # Ensure the token request doesn't work a second time
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 403, r.json()
    assert (
        f"{unallowed_property} are not valid properties for user" in r.json()["detail"]
    )


async def test_flows_with_invalid_properties(test_client):
    """Test the authorization flow and the device flow with invalid properties."""
    invalid_property = "InvalidAndUnknownProperty"

    # Initiate the authorization flow
    code_verifier = secrets.token_hex()
    code_challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode()).digest())
        .decode()
        .replace("=", "")
    )
    r = test_client.get(
        "/api/auth/authorize",
        params={
            "response_type": "code",
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "client_id": DIRAC_CLIENT_ID,
            "redirect_uri": "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
            "scope": f"vo:lhcb property:{invalid_property} property:NormalUser",
            "state": "external-state",
        },
        follow_redirects=False,
    )
    assert r.status_code == 400, r.json()
    assert f"{{'{invalid_property}'}} are not valid properties" in r.json()["detail"]

    # Initiate the device flow
    r = test_client.post(
        "/api/auth/device",
        params={
            "client_id": DIRAC_CLIENT_ID,
            "scope": f"vo:lhcb group:lhcb_user property:{invalid_property} property:NormalUser",
        },
    )
    assert r.status_code == 400, r.json()
    assert f"{{'{invalid_property}'}} are not valid properties" in r.json()["detail"]


async def test_refresh_token_rotation(test_client, auth_httpx_mock: HTTPXMock):
    """Test the refresh token rotation.

    - initiate a device code flow to get an initial refresh token
    - use the refresh token to get a new access token
    - make sure that the initial refresh token is different from the new one (refresh token rotation
    - act as a malicious attacker providing an old refresh token and make sure it has been revoked
    - make sure the user needs to reauthenticate to get a new refresh token
    - last attempt, try to get a refresh token from a non-existing refresh token.
    """
    initial_refresh_token = _get_tokens(test_client)["refresh_token"]

    # ...
    # A malicious attacker manages to steal the initial refresh token of the user
    # ...

    # Malicious attacker gets a new refresh token (to get an access token)
    request_data = {
        "grant_type": "refresh_token",
        "refresh_token": initial_refresh_token,
        "client_id": DIRAC_CLIENT_ID,
    }
    response_data = _get_and_check_token_response(
        test_client, request_data=request_data
    )
    new_refresh_token = response_data["refresh_token"]

    # Make sure it is different from the initial refresh token
    assert initial_refresh_token != new_refresh_token

    # ...
    # User is not aware of the malicious attack
    # User works with an access token until expiration, then needs to get a new one
    # ...

    # User uses the initial refresh token to get a new one
    # The server should detect the breach and revoke every token bound to User
    r = test_client.post("/api/auth/token", data=request_data)
    data = r.json()
    assert r.status_code == 401, data
    assert (
        data["detail"]
        == "Revoked refresh token reused: potential attack detected. You must authenticate again"
    )

    # Make sure that Malicious attacker cannot get a new refresh token from the latest refresh token obtained
    # In theory, new_refresh_token has not been revoked since it is the latest one
    # But because a breach was detected, it should also be revoked
    request_data["refresh_token"] = new_refresh_token
    r = test_client.post("/api/auth/token", data=request_data)
    data = r.json()
    assert r.status_code == 401, data
    assert (
        data["detail"]
        == "Revoked refresh token reused: potential attack detected. You must authenticate again"
    )


async def test_refresh_token_expired(
    test_client, test_auth_settings: AuthSettings, auth_httpx_mock: HTTPXMock
):
    """Test the expiration date of the passed refresh token.
    - get a refresh token
    - decode it and change the expiration time
    - recode it (with the JWK of the server).
    """
    # Get refresh token
    initial_refresh_token = _get_tokens(test_client)["refresh_token"]

    # Decode it
    refresh_payload = jwt.decode(
        initial_refresh_token, options={"verify_signature": False}
    )

    # Modify the expiration time (utc now - 5 hours)
    refresh_payload["exp"] = int(
        (datetime.now(tz=timezone.utc) - timedelta(hours=5)).timestamp()
    )

    # Encode it differently
    new_refresh_token = create_token(refresh_payload, test_auth_settings)

    request_data = {
        "grant_type": "refresh_token",
        "refresh_token": new_refresh_token,
        "client_id": DIRAC_CLIENT_ID,
    }

    # Try to get a new access token using the invalid refresh token
    # The server should detect that it is not encoded properly
    r = test_client.post("/api/auth/token", data=request_data)
    data = r.json()
    assert r.status_code == 401, data
    assert data["detail"] == "expired_token: The token is expired"


async def test_access_token_expired(
    test_client, test_auth_settings: AuthSettings, auth_httpx_mock: HTTPXMock
):
    """Test the expiration date of the passed access token.
    - get an access token
    - decode it and change the expiration time
    - recode it (with the JWK of the server).
    """
    # Get access token
    initial_access_token = _get_tokens(test_client)["access_token"]

    # Decode it
    access_payload = jwt.decode(
        initial_access_token, options={"verify_signature": False}
    )

    # Modify the expiration time (utc now - 5 hours)
    access_payload["exp"] = int(
        (datetime.now(tz=timezone.utc) - timedelta(hours=5)).timestamp()
    )

    # Encode it differently
    new_access_token = create_token(access_payload, test_auth_settings)

    headers = {"Authorization": f"Bearer {new_access_token}"}

    # Try to get the userinfo using the invalid access token
    # The server should detect that it is not encoded properly
    r = test_client.get("/api/auth/userinfo", headers=headers)
    data = r.json()
    assert r.status_code == 401, data
    assert data["detail"] == "Invalid JWT"


async def test_invalid_access_token_signature(
    test_client, test_auth_settings: AuthSettings, auth_httpx_mock: HTTPXMock
):
    """Test the signature of a token.
    - get an access token
    - decode it and change the signature
    - recode it (with the JWK of the server).
    """
    # Get access token
    initial_access_token = _get_tokens(test_client)["access_token"]

    # Decode it
    access_payload = jwt.decode(
        initial_access_token, options={"verify_signature": False}
    )

    # Encode it differently
    new_access_token = create_token(access_payload, test_auth_settings)

    # JWT tokens are structured as: header.payload.signature
    parts = new_access_token.split(".")
    parts[2] = "invalidsig"  # Replace the signature with garbage
    new_access_token = ".".join(parts)

    headers = {"Authorization": f"Bearer {new_access_token}"}

    # Try to get the userinfo using the invalid access token
    # The server should detect that it is not encoded properly
    r = test_client.get("/api/auth/userinfo", headers=headers)
    data = r.json()
    assert r.status_code == 401, data
    assert data["detail"] == "Invalid JWT"


async def test_invalid_access_token_alg_none(
    test_client, test_auth_settings: AuthSettings, auth_httpx_mock: HTTPXMock
):
    """Test token with 'alg': 'none' header (alg:none attack).
    - get a valid token
    - decode payload
    - craft a JWT with alg=none and no signature
    - verify that the server rejects it.
    """
    # Get a valid token
    initial_access_token = _get_tokens(test_client)["access_token"]

    # Decode payload (skip verifying signature)
    access_payload = jwt.decode(
        initial_access_token, options={"verify_signature": False}
    )

    # Manually build JWT with alg=none (no signature)
    # Compared to test_invalid_access_token_signature, the signature is adapted to the alg
    header = {"alg": "none", "typ": "JWT"}
    encoded_header = (
        base64.urlsafe_b64encode(json.dumps(header).encode()).rstrip(b"=").decode()
    )
    encoded_payload = (
        base64.urlsafe_b64encode(json.dumps(access_payload).encode())
        .rstrip(b"=")
        .decode()
    )
    tampered_token = f"{encoded_header}.{encoded_payload}."  # No signature

    headers = {"Authorization": f"Bearer {tampered_token}"}

    # Expect 401 due to disabled alg=none
    r = test_client.get("/api/auth/userinfo", headers=headers)
    data = r.json()
    assert r.status_code == 401, data
    assert data["detail"] == "Invalid JWT"


async def test_refresh_token_rotated_expiration_time(
    test_client, test_auth_settings: AuthSettings, auth_httpx_mock: HTTPXMock
):
    """Test the expiration date of the newly generated refresh token is similar to the previous one.
    - get a refresh token
    - decode it and change the expiration time
    - recode it (with the JWK of the server)
    - get a new refresh token using the old one
    - check that the new refresh token is not expired but has a similar expiration time.
    """
    # Get refresh token
    initial_refresh_token = _get_tokens(test_client)["refresh_token"]

    # Decode it
    refresh_payload = jwt.decode(
        initial_refresh_token, options={"verify_signature": False}
    )

    # Modify the expiration time (utc now + 5 hours)
    refresh_payload["exp"] = int(
        (datetime.now(tz=timezone.utc) + timedelta(hours=5)).timestamp()
    )

    # Encode it differently
    new_refresh_token = create_token(refresh_payload, test_auth_settings)

    request_data = {
        "grant_type": "refresh_token",
        "refresh_token": new_refresh_token,
        "client_id": DIRAC_CLIENT_ID,
    }

    # Try to get a new access token using the invalid refresh token
    # The server should detect that it is not encoded properly
    r = test_client.post("/api/auth/token", data=request_data)
    data = r.json()
    assert r.status_code == 200, data

    # Check that the new refresh token expiration time is similar to the previous one (modified)
    new_refresh_payload = jwt.decode(
        data["refresh_token"], options={"verify_signature": False}
    )
    assert abs(new_refresh_payload["exp"] - refresh_payload["exp"]) <= 2


async def test_refresh_token_invalid(test_client, auth_httpx_mock: HTTPXMock):
    """Test the validity of the passed refresh token.
    - get a refresh token
    - decode it and recode it with a different JWK key.
    """
    # Get refresh token
    initial_refresh_token = _get_tokens(test_client)["refresh_token"]

    # Decode it
    refresh_payload = jwt.decode(
        initial_refresh_token, options={"verify_signature": False}
    )

    # Encode it differently (using another algorithm)
    key = RSAKey.generate_key(
        2048,
        {
            "key_ops": ["sign", "verify"],
            "alg": "RS256",
            "kid": uuid7().hex,
        },
    )

    new_auth_settings = AuthSettings(
        token_issuer="https://iam-auth.web.cern.ch/",
        token_allowed_algorithms=["EdDSA", "RS256"],
        token_keystore=json.dumps(KeySet(keys=[key]).as_dict(private=True)),
        state_key=Fernet.generate_key(),
        allowed_redirects=[
            "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
        ],
    )
    new_refresh_token = create_token(refresh_payload, new_auth_settings)

    # Make sure it is different from the initial refresh token
    assert initial_refresh_token != new_refresh_token

    request_data = {
        "grant_type": "refresh_token",
        "refresh_token": new_refresh_token,
        "client_id": DIRAC_CLIENT_ID,
    }

    # Try to get a new access token using the invalid refresh token
    # The server should detect that it is not encoded properly
    r = test_client.post("/api/auth/token", data=request_data)
    data = r.json()
    assert r.status_code == 401, data
    assert "No key for kid" in data["detail"]


async def test_keystore(test_client):
    """Test the keystore."""
    # Basic Auth params
    issuer = "https://iam-auth.web.cern.ch/"
    state_key = Fernet.generate_key()
    allowed_redirects = [
        "http://diracx.test.invalid:8000/api/docs/oauth2-redirect",
    ]

    payload = {
        "jti": "49ecc171-20be-5b88-0d65-26012c07f397",
        "exp": (datetime.now(tz=timezone.utc) + timedelta(hours=1)).timestamp(),
        "legacy_exchange": False,
    }

    # Generate keys
    rsa_key = RSAKey.generate_key(
        2048,
        {
            "key_ops": ["sign", "verify"],
            "alg": "RS256",
            "kid": uuid7().hex,
        },
    )
    eddsa_key = OKPKey.generate_key(
        "Ed25519",
        {
            "key_ops": ["sign", "verify"],
            "alg": "EdDSA",
            "kid": uuid7().hex,
        },
    )

    # Generate the keystore with eddsa key only first
    jwks = KeySet(keys=[eddsa_key])

    # Generate the keystore with rsa key only first
    auth_settings = AuthSettings(
        token_issuer=issuer,
        token_allowed_algorithms=["RS256"],  # We purposefully remove EdDSA
        token_keystore=json.dumps(jwks.as_dict(private=True)),
        state_key=state_key,
        allowed_redirects=allowed_redirects,
    )

    # Encode/Decode with the keystore: should not work
    # because EdDSA is not part of the allowed algorithms
    with pytest.raises(UnsupportedAlgorithmError):
        token = create_token(payload, auth_settings)

    # Add EdDSA to the allowed algorithms
    auth_settings.token_allowed_algorithms.append("EdDSA")

    # Encode/Decode with the keystore: should work
    token = create_token(payload, auth_settings)
    await verify_dirac_refresh_token(token, auth_settings)

    # Add the rsa key to the keystore
    jwks.keys.append(rsa_key)

    auth_settings = AuthSettings(
        token_issuer=issuer,
        token_allowed_algorithms=["RS256", "EdDSA"],  # We purposefully remove EdDSA
        token_keystore=json.dumps(jwks.as_dict(private=True)),
        state_key=state_key,
        allowed_redirects=allowed_redirects,
    )

    # Encode/Decode with the keystore: should work
    token = create_token(payload, auth_settings)
    await verify_dirac_refresh_token(token, auth_settings)

    # Remove 'sign' operation from the RSA key:
    # should still work because eddsa_key is still there
    auth_settings.token_keystore.jwks.keys[1].get("key_ops").remove("sign")
    token = create_token(payload, auth_settings)
    await verify_dirac_refresh_token(token, auth_settings)

    # Remove 'verify' operation from the RSA key:
    # should still work because RSA key is still in the keystore
    auth_settings.token_keystore.jwks.keys[1].get("key_ops").remove("verify")
    token = create_token(payload, auth_settings)
    await verify_dirac_refresh_token(token, auth_settings)

    # Remove RSA key from the keystore:
    # should still work except decoding the token signed with the RSA key
    auth_settings.token_keystore.jwks.keys.pop(1)
    ed_signed_token = create_token(payload, auth_settings)
    await verify_dirac_refresh_token(ed_signed_token, auth_settings)

    # Remove 'sign' operation from the OPK key:
    # Should not work
    auth_settings.token_keystore.jwks.keys[0].get("key_ops").remove("sign")

    with pytest.raises(ValueError):
        create_token(payload, auth_settings)

    # But we can still verify the token
    await verify_dirac_refresh_token(ed_signed_token, auth_settings)

    # Remove 'verify' operation from the OPK key:
    # Verification should still not work anymore
    auth_settings.token_keystore.jwks.keys[0].get("key_ops").remove("verify")

    with pytest.raises(ValueError):
        # This should raise an error because the key is not usable for verifying
        create_token(payload, auth_settings)

    with pytest.raises(UnsupportedKeyOperationError):
        await verify_dirac_refresh_token(ed_signed_token, auth_settings)

    # Remove the key from the keystore:
    auth_settings.token_keystore.jwks.keys.pop(0)

    with pytest.raises(ValueError):
        # This should raise an error because there is no key in the keystore
        create_token(payload, auth_settings)

    # This should raise an error because there is no key in the keystore
    with pytest.raises(InvalidKeyIdError):
        await verify_dirac_refresh_token(ed_signed_token, auth_settings)


async def test_bad_access_token(test_client):
    """Test accessing a resource with a bad token."""
    # From https://github.com/DIRACGrid/diracx/pull/496
    r = test_client.get(
        "/api/auth/userinfo", headers={"Authorization": "Bearer thisisabadbearer"}
    )
    data = r.json()

    # Should raise a 401, with "Invalid JWT"
    # Not Invalid Authorization Header because raised when decoding the token
    assert r.status_code == 401, data
    assert data["detail"] == "Invalid JWT"


async def test_list_refresh_tokens(test_client, auth_httpx_mock: HTTPXMock):
    """Test the refresh token listing with 2 users, a normal one and token manager:
    - normal user gets a refresh token and lists it
    - token manager gets a refresh token and lists all of them
    - normal user renews his/her refresh token and list it: should have only one as the first one should be revoked
    - token manager lists all of them: should still see it as revoked.
    """
    # Normal user gets a pair of tokens
    normal_user_tokens = _get_tokens(test_client, property=NORMAL_USER)
    normal_user_access_token = normal_user_tokens["access_token"]
    normal_user_refresh_token = normal_user_tokens["refresh_token"]

    # Normal user lists his/her refresh tokens
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data
    assert len(data) == 1

    # token manager gets a pair of tokens
    token_manager_access_token = _get_tokens(
        test_client, group="lhcb_tokenmgr", property=PROXY_MANAGEMENT
    )["access_token"]

    # Token manager lists refresh tokens: should get his/her own and the normal user's one
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {token_manager_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data
    assert len(data) == 2

    # Normal user gets a new refresh token
    request_data = {
        "grant_type": "refresh_token",
        "refresh_token": normal_user_refresh_token,
        "client_id": DIRAC_CLIENT_ID,
    }
    response_data = _get_and_check_token_response(
        test_client, request_data=request_data
    )

    # Normal user lists his/her refresh tokens again
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {response_data['access_token']}"},
    )
    data = r.json()
    assert r.status_code == 200, data
    assert len(data) == 1

    # Token manager lists refresh tokens: should get his/her own and the normal user's one
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {token_manager_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data
    assert len(data) == 3


async def test_revoke_refresh_tokens_normal_user_with_jti(
    test_client, auth_httpx_mock: HTTPXMock
):
    """Test the refresh token revokation with 2 users, a normal one and token manager:
    - normal user gets a refresh token
    - token manager gets a refresh token
    - normal user tries to delete a non-existing RT: should not work
    - normal user tries to delete the token manager's RT: should not work
    - normal user tries to delete his/her RT: should work.
    """
    # Normal user gets a pair of tokens
    normal_user_tokens = _get_tokens(test_client, property=NORMAL_USER)
    normal_user_access_token = normal_user_tokens["access_token"]
    normal_user_refresh_token = normal_user_tokens["refresh_token"]
    normal_user_refresh_payload = jwt.decode(
        normal_user_refresh_token, options={"verify_signature": False}
    )

    # Token manager gets a pair of tokens
    token_manager_tokens = _get_tokens(
        test_client, group="lhcb_tokenmgr", property=PROXY_MANAGEMENT
    )
    token_manager_refresh_token = token_manager_tokens["refresh_token"]
    token_manager_refresh_payload = jwt.decode(
        token_manager_refresh_token, options={"verify_signature": False}
    )

    # Normal user tries to delete a random and non-existing RT: should raise an error
    r = test_client.delete(
        "/api/auth/refresh-tokens/does-not-exists",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )
    data = r.json()
    assert r.status_code == 400, data

    # Normal user tries to delete token manager's RT: should not work
    r = test_client.delete(
        f"/api/auth/refresh-tokens/{token_manager_refresh_payload['jti']}",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )
    data = r.json()
    assert r.status_code == 403, data

    # Normal user tries to delete his/her RT: should work
    r = test_client.delete(
        f"/api/auth/refresh-tokens/{normal_user_refresh_payload['jti']}",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data

    # Normal user tries to delete his/her RT again: should work
    r = test_client.delete(
        f"/api/auth/refresh-tokens/{normal_user_refresh_payload['jti']}",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data


async def test_revoke_refresh_tokens_token_manager_with_jti(
    test_client, auth_httpx_mock: HTTPXMock
):
    """Test the refresh token revokation with 2 users, a normal one and token manager:
    - normal user gets a refresh token
    - token manager gets a refresh token
    - token manager tries to delete normal user's RT: should work
    - token manager tries to delete his/her RT: should work too.
    """
    # Normal user gets a pair of tokens
    normal_user_tokens = _get_tokens(test_client, property=NORMAL_USER)
    normal_user_refresh_token = normal_user_tokens["refresh_token"]
    normal_user_refresh_payload = jwt.decode(
        normal_user_refresh_token, options={"verify_signature": False}
    )

    # Token manager gets a pair of tokens
    token_manager_tokens = _get_tokens(
        test_client, group="lhcb_tokenmgr", property=PROXY_MANAGEMENT
    )
    token_manager_access_token = token_manager_tokens["access_token"]
    token_manager_refresh_token = token_manager_tokens["refresh_token"]
    token_manager_refresh_payload = jwt.decode(
        token_manager_refresh_token, options={"verify_signature": False}
    )

    # Token manager tries to delete token manager's RT: should work
    r = test_client.delete(
        f"/api/auth/refresh-tokens/{normal_user_refresh_payload['jti']}",
        headers={"Authorization": f"Bearer {token_manager_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data

    # Token manager tries to delete his/her RT: should work
    r = test_client.delete(
        f"/api/auth/refresh-tokens/{token_manager_refresh_payload['jti']}",
        headers={"Authorization": f"Bearer {token_manager_access_token}"},
    )
    data = r.json()
    assert r.status_code == 200, data


async def test_revoke_refresh_token_classic(test_client, auth_httpx_mock: HTTPXMock):
    """Test if a user can revoke its token by the "traditional" endpoint, then try to refresh it."""
    # Normal user gets a pair of tokens
    normal_user_tokens = _get_tokens(test_client, property=NORMAL_USER)
    normal_user_access_token = normal_user_tokens["access_token"]
    normal_user_refresh_token = normal_user_tokens["refresh_token"]

    # Get all refresh tokens, expect one refresh token
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )

    assert len(r.json()) == 1

    # See: https://datatracker.ietf.org/doc/html/rfc7009#section-2.2
    # Normal user tries to delete a random and non-existing RT: should respond with a 200
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": "does-not-exist",
            "client_id": DIRAC_CLIENT_ID,
        },
    )
    assert r.status_code == 200

    # Normal user tries to delete his/her RT: should work
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": normal_user_refresh_token,
            "client_id": DIRAC_CLIENT_ID,
        },
    )
    assert r.status_code == 200

    # Get all refresh tokens, expect no refresh token
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )

    assert r.json() == []

    # Normal user tries to delete a valid RT using the wrong client id
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": normal_user_refresh_token,
            "client_id": "a_wrong_dirac_client_id",
        },
    )
    assert r.status_code == 401
    assert r.json()["detail"] == "Unrecognised client_id"

    # Normal user tries to delete a valid access token: should not work but respond with a 200
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": normal_user_access_token,
            "client_id": DIRAC_CLIENT_ID,
        },
    )
    assert r.status_code == 200

    # Get all refresh tokens, expect no refresh token
    r = test_client.get(
        "/api/auth/refresh-tokens",
        headers={"Authorization": f"Bearer {normal_user_access_token}"},
    )

    assert r.json() == []

    # Normal user tries to delete a valid access token by specifying the token_type_hint: should not work
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": normal_user_access_token,
            "client_id": DIRAC_CLIENT_ID,
            "token_type_hint": "access_token",
        },
    )
    assert r.status_code == 400
    assert r.json()["detail"] == "unsupported_token_type"

    # Normal user tries to delete a valid access token by specifying an invalid token_type_hint:
    # Should not work but respond with a 200
    r = test_client.post(
        "/api/auth/revoke",
        data={
            "token": normal_user_access_token,
            "client_id": DIRAC_CLIENT_ID,
            "token_type_hint": "invalid_token_type",
        },
    )
    assert r.status_code == 200


def _get_tokens(
    test_client, group: str = "lhcb_user", property: SecurityProperty = NORMAL_USER
):
    """Get a pair of tokens (access, refresh) through a device flow code."""
    # User Initiates a device flow (would normally be done from CLI)
    r = test_client.post(
        "/api/auth/device",
        params={
            "client_id": DIRAC_CLIENT_ID,
            "scope": f"vo:lhcb group:{group} property:{property}",
        },
    )
    data = r.json()

    # Open the DIRAC login page and ensure it redirects to the IdP
    r = test_client.get(data["verification_uri_complete"], follow_redirects=False)

    login_url = r.headers["Location"]
    query_parameters = parse_qs(urlparse(login_url).query)
    redirect_uri = query_parameters["redirect_uri"][0]
    state = query_parameters["state"][0]

    r = test_client.get(redirect_uri, params={"code": "valid-code", "state": state})

    # User gets a TokenResponse: should contain an access and a refresh tokens
    r = test_client.post(
        "/api/auth/token",
        data={
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": data["device_code"],
            "client_id": DIRAC_CLIENT_ID,
        },
    )
    return r.json()


def _get_and_check_token_response(test_client, request_data):
    """Get a token and check that mandatory fields are present and that the userinfo endpoint returns
    something sensible.
    """
    # Check that token request now works
    r = test_client.post("/api/auth/token", data=request_data)
    assert r.status_code == 200, r.json()
    response_data = r.json()
    assert response_data["access_token"]
    assert response_data["refresh_token"]
    assert response_data["expires_in"]
    assert response_data["token_type"]

    r = test_client.get(
        "/api/auth/userinfo",
        headers={"authorization": f"Bearer {response_data['access_token']}"},
    )
    assert r.status_code == 200, r.json()

    return response_data


@pytest.mark.parametrize(
    "vos, groups, scope, expected",
    [
        # We ask for a vo, we get the properties of the default group
        [
            {"lhcb": {"default_group": "lhcb_user"}},
            {
                "lhcb_user": {"properties": ["NormalUser"]},
                "lhcb_admin": {"properties": ["ProxyManagement"]},
                "lhcb_production": {"properties": ["ProductionManagement"]},
            },
            "vo:lhcb",
            {"group": "lhcb_user", "properties": {"NormalUser"}, "vo": "lhcb"},
        ],
        # We ask for a vo and a group, we get the properties of the group
        [
            {"lhcb": {"default_group": "lhcb_user"}},
            {
                "lhcb_user": {"properties": ["NormalUser"]},
                "lhcb_admin": {"properties": ["ProxyManagement"]},
                "lhcb_production": {"properties": ["ProductionManagement"]},
            },
            "vo:lhcb group:lhcb_admin",
            {"group": "lhcb_admin", "properties": {"ProxyManagement"}, "vo": "lhcb"},
        ],
        # We ask for a vo, no group, and an additional existing property
        # We get the default group with its properties along with with the extra properties we asked for
        # Authorization to access the additional property is checked later when user effectively requests a token
        [
            {"lhcb": {"default_group": "lhcb_user"}},
            {
                "lhcb_user": {"properties": ["NormalUser"]},
                "lhcb_admin": {"properties": ["ProxyManagement"]},
                "lhcb_production": {"properties": ["ProductionManagement"]},
            },
            "vo:lhcb property:ProxyManagement",
            {
                "group": "lhcb_user",
                "properties": {"NormalUser", "ProxyManagement"},
                "vo": "lhcb",
            },
        ],
        # We ask for a vo and a group with additional property
        # We get the properties of the group + the additional property
        # Authorization to access the additional property is checked later when user effectively requests a token
        [
            {"lhcb": {"default_group": "lhcb_user"}},
            {
                "lhcb_user": {"properties": ["NormalUser"]},
                "lhcb_admin": {"properties": ["ProxyManagement"]},
                "lhcb_production": {"properties": ["ProductionManagement"]},
            },
            "vo:lhcb group:lhcb_admin property:ProductionManagement",
            {
                "group": "lhcb_admin",
                "properties": {"ProductionManagement", "ProxyManagement"},
                "vo": "lhcb",
            },
        ],
    ],
)
def test_parse_scopes(vos, groups, scope, expected):
    config = Config.model_validate(
        {
            "DIRAC": {},
            "Registry": {
                vo_name: {
                    "DefaultGroup": vo_conf["default_group"],
                    "IdP": {"URL": "https://idp.invalid", "ClientID": "test-idp"},
                    "Users": {},
                    "Groups": {
                        group_name: {
                            "Properties": group_conf["properties"],
                            "Users": [],
                        }
                        for group_name, group_conf in groups.items()
                    },
                }
                for vo_name, vo_conf in vos.items()
            },
        }
    )
    available_properties = SecurityProperty.available_properties()
    assert parse_and_validate_scope(scope, config, available_properties) == expected


@pytest.mark.parametrize(
    "vos, groups, scope, expected_error",
    [
        [
            ["lhcb"],
            ["lhcb_user"],
            "group:lhcb_user undefinedscope:undefined",
            "Unrecognised scopes",
        ],
        [
            ["lhcb"],
            ["lhcb_user", "lhcb_admin"],
            "vo:lhcb group:lhcb_user property:undefined_property",
            "{'undefined_property'} are not valid properties",
        ],
        [
            ["lhcb"],
            ["lhcb_user"],
            "group:lhcb_user",
            "No vo scope requested",
        ],
        [
            ["lhcb", "gridpp"],
            ["lhcb_user", "lhcb_admin"],
            "vo:lhcb vo:gridpp group:lhcb_user group:lhcb_admin",
            "Only one vo is allowed",
        ],
        [
            ["lhcb"],
            ["lhcb_user", "lhcb_admin"],
            "vo:lhcb group:lhcb_user group:lhcb_admin",
            "Only one DIRAC group allowed",
        ],
    ],
)
def test_parse_scopes_invalid(vos, groups, scope, expected_error):
    config = Config.model_validate(
        {
            "DIRAC": {},
            "Registry": {
                vo: {
                    "DefaultGroup": "lhcb_user",
                    "IdP": {"URL": "https://idp.invalid", "ClientID": "test-idp"},
                    "Users": {},
                    "Groups": {
                        group: {"Properties": ["NormalUser"], "Users": []}
                        for group in groups
                    },
                }
                for vo in vos
            },
            "Operations": {},
        }
    )
    available_properties = SecurityProperty.available_properties()
    with pytest.raises(ValueError, match=expected_error):
        parse_and_validate_scope(scope, config, available_properties)


def test_encrypt_decrypt_state_valid_state(fernet_key):
    """Test that decrypt_state returns the correct state."""
    fernet = Fernet(fernet_key)
    # Create a valid state
    state_dict = {
        "vo": "lhcb",
        "code_verifier": secrets.token_hex(),
        "user_code": "AE19U",
        "grant_type": GrantType.device_code.value,
    }

    state = encrypt_state(state_dict, fernet)
    result = decrypt_state(state, fernet)

    assert result == state_dict

    # Create an empty state
    state_dict = {}

    state = encrypt_state(state_dict, fernet)
    result = decrypt_state(state, fernet)

    assert result == state_dict


def test_encrypt_decrypt_state_invalid_state(fernet_key):
    """Test that decrypt_state raises an error when the state is invalid."""
    state = "invalid_state"  # Invalid state string

    with pytest.raises(AuthorizationError) as exc_info:
        decrypt_state(state, fernet_key)
    assert exc_info.value.detail == "Invalid state"



from __future__ import annotations

import asyncio
from functools import partial

import pytest
from git import Repo

from diracx.core.config import ConfigSource

pytestmark = pytest.mark.enabled_dependencies(["AuthDB", "ConfigSource"])


async def test_live(client_factory):
    config_source = client_factory.all_dependency_overrides[
        ConfigSource.create
    ].__self__

    with client_factory.unauthenticated() as client:
        probe_fcn = partial(client.get, "/api/health/live")
        r = await _test_after_clear(config_source, probe_fcn)
        assert r.json() == {"status": "live"}

        old_config = await config_source.read_config_non_blocking()
        repo = Repo(config_source.repo_location)
        commit = repo.index.commit("Test commit")
        assert commit.hexsha not in old_config._hexsha

        # This will trigger a config update but still return the old config
        config_source._revision_cache.soft_cache.clear()
        new_config = await config_source.read_config_non_blocking()
        assert commit.hexsha not in old_config._hexsha

        # If we wait long enough, the new config should be returned
        for _ in range(50):
            await asyncio.sleep(0.1)

            r = probe_fcn()
            assert r.status_code == 200, r.text
            new_config = await config_source.read_config_non_blocking()
            if commit.hexsha in new_config._hexsha:
                break
        else:
            raise AssertionError("New config not found in probe response", r.text)

        # Ensure everything is still okay
        r = probe_fcn()
        assert r.status_code == 200, r.text


async def test_ready(client_factory):
    config_source = client_factory.all_dependency_overrides[
        ConfigSource.create
    ].__self__

    with client_factory.unauthenticated() as client:
        probe_fcn = partial(client.get, "/api/health/ready")
        r = await _test_after_clear(config_source, probe_fcn)
        assert r.json() == {"status": "ready"}


async def test_startup(client_factory):
    config_source = client_factory.all_dependency_overrides[
        ConfigSource.create
    ].__self__

    with client_factory.unauthenticated() as client:
        probe_fcn = partial(client.get, "/api/health/startup")
        r = await _test_after_clear(config_source, probe_fcn)
        assert r.json() == {"status": "startup complete"}


async def _test_after_clear(config_source, probe_fcn):
    """Ensure that the probe fails after clearing the config source caches.

    :param config_source: The config source to clear.
    :param probe_fcn: The function to call to make the probe request.

    :return: The response from the probe.
    """
    orig_r = probe_fcn()
    assert orig_r.status_code == 200, orig_r.text

    # Ensure that a 503 error is returned if the config source is not ready
    config_source.clear_caches()
    r = probe_fcn()
    assert r.status_code == 503, r.text

    # Ensure that the startup probe returns to normal after a short delay
    for _ in range(50):
        r = probe_fcn()
        if r.status_code != 503:
            assert r.status_code == 200, r.text
            assert r.json() == orig_r.json(), (r.text, orig_r.text)
            return r
        await asyncio.sleep(0.1)
    else:
        raise AssertionError("Probe did not return to normal", r.text)
